#!/usr/bin/env python

################################################################################
#
# Poly
# Version 0.0.7
#
# Copyright (c) 1996-2001 J.W. Bizzaro & University of Massachusetts Lowell
#
# Program for the quantitative analysis of simple sequence repeats (SSRs)
# in DNA
#
# ------------
#
# Poly is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
################################################################################

import sys, string
from math import log

FALSE = 0
TRUE = 1
MIN_COUNT = 2
HASH_MARKS = 10.0


class Polymer:

	def __init__(self):
		"""
		"""
	
		self.name = ''
		self.monomer_count = 0
		self.count = 0


class PolyList:

	def __init__(self):
		"""
		"""

		# initialize list
		self.polymer = []



class Poly:

	def __init__(self):
		"""
		"""

		start = 'yes'
		self.infiles = []
		self.monomerlength = 1
		self.infilename = ''

		index = 1
		for option in sys.argv[1:]:

			if (option == '-l') | (option == '--length'):
				self.monomerlength = int(sys.argv[index+1])

			elif (option == '-i') | (option == '--infiles'):
				self.infiles = sys.argv[(index + 1):]

			elif (option == '-h') | (option == '--help'):
				print 'Please see the README file.'
				start = 'no'

			elif (option == '-v') | (option == '--version'):
				print 'Version 0.0.7'
				start = 'no'

			index = index + 1

		if len(self.infiles) == 0:
			print 'no infiles'
			start = 'no'

		if start == 'yes':

			# Do an analysis for each file in the filename list
			for self.infilename in self.infiles:
				self.infile = open(self.infilename, 'r')

				self.totalbases = 0
				self.basecount = 0
				self.G_count = 0
				self.C_count = 0
				self.A_count = 0
				self.T_count = 0
				self.U_count = 0
				self.Y_count = 0
				self.N_count = 0

				self.percent = 0
				self.mark_printed = FALSE
				self.partial = 0
				self.whole_monos = 1
				self.highestpolynumber = 0

				self.polylist = PolyList()
				self.init_window()

				base = self.nextbase()

				while base != '':
					self.checkbase(base)
					self.addbasetowindow(base)
					base = self.nextbase()

				self.totalbases = self.basecount

				self.A_freq = float(self.A_count) / float(self.totalbases)
				self.C_freq = float(self.C_count) / float(self.totalbases)
				self.G_freq = float(self.G_count) / float(self.totalbases)
				self.T_freq = float(self.T_count) / float(self.totalbases)
				self.U_freq = float(self.U_count) / float(self.totalbases)
				self.Y_freq = float(self.Y_count) / float(self.totalbases)
				self.N_freq = float(self.N_count) / float(self.totalbases)

		#		sys.stdout.write(datetostr(date), ' ', timetostr(time))
				sys.stdout.write('Sequence file: ' + self.infilename + '\n')
				sys.stdout.write('Total bases: ' + str(self.totalbases) + '\n')
				sys.stdout.write('Frequency of A: ' + str(self.A_freq) + '\n')
				sys.stdout.write('Frequency of C: ' + str(self.C_freq) + '\n')
				sys.stdout.write('Frequency of G: ' + str(self.G_freq) + '\n')
				sys.stdout.write('Frequency of T: ' + str(self.T_freq) + '\n')
				sys.stdout.write('Frequency of U: ' + str(self.U_freq) + '\n')
				sys.stdout.write('Frequency of Y: ' + str(self.Y_freq) + '\n')
				sys.stdout.write('Frequency of N: ' + str(self.N_freq) + '\n')
				sys.stdout.write('GC composition: ' + str((self.G_freq + self.C_freq) * 100) + '%\n')
				sys.stdout.write('AT composition: ' + str((self.A_freq + self.T_freq) * 100) + '%\n')

				self.writeallpolys()
				self.infile.close()


	# End __init__


	def init_window(self):
		"""
		Create self.window and read in the first bases to fill self.window
		"""

		self.window = []

		for value in range(0, self.monomerlength):
			base = self.nextbase()
			self.window.append(base)

	# End init_wiindow


	def nextbase(self):
		"""
		Read from infile the next wanted base
		"""


		wanted = FALSE
		while not(wanted):
			base = self.infile.read(1)
			# Convert base to uppercase
			base = string.upper(base)
			wanted = FALSE
			if base == 'A':
				wanted = TRUE
				self.A_count = self.A_count + 1
			elif base == 'C':
				wanted = TRUE
				self.C_count = self.C_count + 1
			elif base == 'G':
				wanted = TRUE
				self.G_count = self.G_count + 1
			elif base == 'T':
				wanted = TRUE
				self.T_count = self.T_count + 1
			elif base == 'U':
				wanted = TRUE
				self.U_count = self.U_count + 1
			elif base == 'Y':
				wanted = TRUE
				self.Y_count = self.Y_count + 1
			elif base == 'N':
				wanted = TRUE
				self.N_count = self.N_count + 1
			elif base == '*':
				wanted = TRUE
			elif base == '':
				wanted = TRUE

		# Some characters are wanted but not counted as bases
		if not((base == '') | (base == '*')):
			self.basecount = self.basecount + 1

# Temp
		if base == '*': base = 'M'

# Remove?
#			if self.totalbases > 0:
#				self.percent = float(self.basecount) \
#				/ float(self.totalbases) * 100.0
#
#				if int((self.percent % (100.0 / HASH_MARKS)) * 10.0) == 0:
#					if not(self.mark_printed):
#						print str(int(self.percent)) + '%'
#						self.mark_printed = TRUE
#				else:
#					self.mark_printed = FALSE

		return base

	# End nextbase


	def checkbase(self, base):
		"""
		"""

		# If next base matches the first base in window,
		# then there might be another monomer
		if base == self.window[0]:
			# We have at least a partial monomer
			self.partial = self.partial + 1

			# If we've reached the monomer length, then we have
			# a whole monomer
			if self.partial == self.monomerlength:
				self.whole_monos = self.whole_monos + 1
				self.partial = 0

		# If next base does NOT match the first base in window
		# then the polymer can be added to the list
		else:
			nested_polymer = FALSE

			# Homopolymers can't be nested polymers
			if self.monomerlength > 1:
				# Check to see if the monomer contains its own polymer
				nested_polymer = self.check_nested_polymer()

			# If the polymer is not a nested polymer
			if not(nested_polymer):
				# Then the polymer can be added to the list
				self.add_a_poly()

			self.partial = 0
			self.whole_monos = 1

	# End checkbase 


	def check_nested_polymer(self):

		"""
		Routine to check for nested polymers (nps).	This is when the
		monomer being examined is actually a homopolymer or a polymer
		of a smaller monomer size.

		Any monomer can contain submonomers of any number into which the monomer
		can be divided.  For example, a monomer of 8 bases can contain 8x 1-mers,
		4x 2-mers, 2x 4-mers (and 1x 8-mer).  If all of the submonomers of a
		given size match, then we can say the monomer is a "subpolymer within
		the entire scope of the monomer"** and not a true monomer.

		**If you had, for example, a monomer length of 8, and a monomer which
		appears as "GAAATCGT", the "AAA" is an A8 repeat but does not fill the
		entire monomer.  Also, if the the monomer appears as "AAAACCCC", there
		is still no repeat within the scope of the monomer.
		"""

		monomer = self.window
		monomerlength = self.monomerlength

		# Submonomer lengths can be from 1 to the monomerlength - 1
		# Starting from smallest lengths, if the smaller lengths are polymeric,
		# then it goes that the larger lengths are too.
		for submonomerlength in range(1, monomerlength):
			# Restrict submonomer lengths to those into which the monomer
			# can be evenly divided
			if (monomerlength % submonomerlength) == 0:

				# Assume it is a subpolymer unless otherwise determined
				is_a_subpolymer = TRUE

				number_of_submonomers = monomerlength / submonomerlength

				# First submonomer is used for reference
				reference_submonomer = monomer[0:submonomerlength]

				# Examine each submonomer of submonomerlength
				for submonomer_number in range(0, number_of_submonomers):

					submonomer_tab = submonomerlength * submonomer_number

					submonomer = \
					monomer[submonomer_tab:submonomer_tab+submonomerlength]

					# Can't be a subpolymer if submonomers don't match
					if submonomer <> reference_submonomer:
						is_a_subpolymer = FALSE

		return is_a_subpolymer

	# End check_nested_polymer


	def add_a_poly(self):
		"""
		Add a polymer to the list
		"""

		# Check if the polymer is already in the list
		polymer = self.find_by_name()
		# If it is not in the list, then add it
		if polymer == None:
			polymer = self.create_a_poly()
		# Increase the number of times this polymer has been observed
		polymer.count = polymer.count + 1


	def find_by_name(self):
		"""
		Check if the polymer's name is in the list of polymers and that
		the number of whole monomers matches.  If not, then add the polymer
		to the list.
		"""

		# Name is simply the sequence of the monomer or window plus the
		# number of whole monomers.  For example, A3 = AAA
		monomer_name = self.window_to_monomer_name()
		polyname = self.polyname(monomer_name, self.whole_monos)

		# Search the list of polymers for one with the same name
		for candidate in self.polylist.polymer:
			if (candidate.name == polyname):
				polymer = candidate
				return polymer

		# If no polymer with same name is found, then return None
		return None


	def polyname(self, monomer_name, monomer_count):

		# Polymer name is simply the monomer name plus the monomer count.
		# For example, GCG03 = GCGGCGGCG
		if self.whole_monos < 10:
			polyname = monomer_name + '0' + str(monomer_count)
		else:
			polyname = monomer_name + str(monomer_count)

		return polyname

	# End polyname


	def create_a_poly(self):
		"""
		Create a new polymer object
		"""

		polymer = Polymer()
		# The name using only bases (no monomer count).  Just the sequence
		# of the monomer or window
		polymer.monomer_name = self.window_to_monomer_name()
		# The number of whole monomers in the polymer
		polymer.monomer_count = self.whole_monos
		# Determine polymer's name
		polymer.name = self.polyname(polymer.monomer_name, \
		polymer.monomer_count)
		# The number of times the polymer occurs.  Increased later.
		polymer.count = 0
		# Add the new polymer to the list
		self.polylist.polymer.append(polymer)

		return polymer

	# End create_a_poly


	def window_to_monomer_name(self):
		"""
		Convert a list into a single string
		"""

		monomer_name = ''
		for base in self.window:
			monomer_name = monomer_name + base

		return monomer_name

	# End window_to_monomer_name 


	def addbasetowindow(self, base):
		"""
		"""

		# Shift bases in self.window one place to the left
		for value in range(0, len(self.window)-1):
			self.window[value] = self.window[value+1]

		self.window[len(self.window)-1] = base

	# End addbasetowindow 


	def writeallpolys(self):
		"""
		"""

		# Create a list to hold sorted polymer names
		polynames_sorted = []

		# Add polymer names to the list
		for polymer in self.polylist.polymer:
			polynames_sorted.append(polymer.name)

		# Sort the list (alpha-numerically)
		polynames_sorted.sort()

		# Create a list to hold sorted polymers
		polylist_sorted = []

		# Put the actual polymers in the order of polymer names
		for polyname in polynames_sorted:
			# For each polymer name, search the polymer list for it
			for polymer in self.polylist.polymer:
				# If found
				if polymer.name == polyname:
					# Add the polymer to the list of sorted polymers
					polylist_sorted.append(polymer)

		# Remove non-polymers (monomer count == 1) from the list
		# ONLY if there are no polymers by the same monomer name
		index = 0
		# Search the list
		for polymer in polylist_sorted:

			# If the polymer is not really a polymer (monomer count == 1)
			if polymer.monomer_count == 1:
				# If there is at least one more polymer in list
				if index < (len(polylist_sorted) - 1):
					# Then check the next polymer in list
					# If it does NOT have the same monomer name
					if polymer.monomer_name \
					!= polylist_sorted[index + 1].monomer_name:
						# Then remove it
						del polylist_sorted[index]
				# If the non-polymer is the last in the list
				else:
						# Then remove it
						del polylist_sorted[index]
			index = index + 1

		index = 0
		# Search the list
		for polymer in polylist_sorted:
			# Divide the print out by monomer name
			# If first polymer
			if index == 0:
				# Then open a new outfile
				outfilename = self.infilename + '.' + polymer.monomer_name
				self.open_outfile(outfilename)
				self.zero_three_dep = FALSE
				self.zero_five_dep = FALSE
				freq_monomer_pred = self.get_freq_monomer_pred(polymer)
				extinction = self.get_extinction(freq_monomer_pred)
				if extinction != -1:
					sys.stdout.write('Extinction for ' \
					+ str(polymer.monomer_name) + ': ' \
					+ str(extinction) + '\n')
				else:
					sys.stdout.write('Div by 0 calculating extinction for ' \
					+ str(polymer.monomer_name) + '\n')

			# If a new monomer name is found
			elif polymer.monomer_name \
			!= polylist_sorted[index - 1].monomer_name:
				# Then close previous outfile...
				self.close_outfile()

				# And open a new one
				outfilename = self.infilename + '.' + polymer.monomer_name
				self.open_outfile(outfilename)
				self.zero_three_dep = FALSE
				self.zero_five_dep = FALSE
				freq_monomer_pred = self.get_freq_monomer_pred(polymer)
				extinction = self.get_extinction(freq_monomer_pred)
				if extinction != -1:
					sys.stdout.write('Extinction for ' \
					+ str(polymer.monomer_name) + ': ' \
					+ str(extinction) + '\n')
				else:
					sys.stdout.write('Div by 0 calculating extinction for ' \
					+ str(polymer.monomer_name) + '\n')

			# Print out the details for polymer
			self.print_details(polymer, freq_monomer_pred)
			index = index + 1

		self.close_outfile()

	# End writeallpolys 


	def open_outfile(self, outfilename):
		"""
		"""
		self.outfile = open(outfilename, 'w')

	# End open_outfile


	def get_freq_monomer_pred(self, polymer):
		"""
		"""

		# Find the predicted frequency of the monomer's occurance.
		# This is found by multiplying the individual OBSERVED base frequencies.
		# This value is used to determine the departure of frequencies
		# from predicted.
		freq_monomer_pred = 1.0
		for base in polymer.monomer_name:
			if base == 'A':
				freq_monomer_pred = freq_monomer_pred * self.A_freq
			elif base == 'C':
				freq_monomer_pred = freq_monomer_pred * self.C_freq
			elif base == 'G':
				freq_monomer_pred = freq_monomer_pred * self.G_freq
			elif base == 'T':
				freq_monomer_pred = freq_monomer_pred * self.T_freq
			elif base == 'U':
				freq_monomer_pred = freq_monomer_pred * self.U_freq
			elif base == 'Y':
				freq_monomer_pred = freq_monomer_pred * self.Y_freq
			elif base == 'N':
				freq_monomer_pred = freq_monomer_pred * self.N_freq

		return freq_monomer_pred

	# End get_freq_monomer_pred

	def close_outfile(self):
		"""
		"""

		self.outfile.close()

	# End close_outfile


	def get_extinction(self, freq_monomer_pred):
		"""
		"""

		# extinction : log (1 / total_bases) / log (freq_monomer_pred)

		if log(freq_monomer_pred) != 0:
			extinction = int(log(1 / float(self.totalbases)) \
			/ log(freq_monomer_pred))

		else:
			extinction = -1

		return extinction

	# End get_extinction


	def print_details(self, polymer, freq_monomer_pred):
		"""
		"""

# freq_polymer_obs     : polymer_count / total_bases
# log_freq_polymer_obs : log(freq_polymer_obs)
# freq_monomer_pred    : freq_base1_obs * freq_base2_obs * ...
#                        This is set when the polymer object is created
# freq_polymer_pred    :
#    freq_monomer_pred^monomer_count * (1 - freq_monomer_pred)^2
#                                           ^^^^^^^^^^^^^
#                           A whole polymer is surrounded by non-monomers
#
# depart_from_pred     : freq_polymer_obs / freq_polymer_pred
# log_depart_from_pred : log(depart_from_pred)

#		polymer_count_max = int((self.totalbases - 1) / (polymer.monomer_count + 1))
		freq_polymer_obs = float(polymer.count) / float(self.totalbases)

		# Note that log() in Python is the natural log or ln()
		log_freq_polymer_obs = log(freq_polymer_obs) / log(10)

		freq_polymer_pred = freq_monomer_pred ** polymer.monomer_count \
		* (1 - freq_monomer_pred ) ** 2

		if freq_polymer_pred != 0:
			depart_from_pred = freq_polymer_obs / freq_polymer_pred
			log_depart_from_pred = log(depart_from_pred) / log(10)

			# Don't print details of polymer if polymer appears less than
			# the desired number of times
			if polymer.count >= MIN_COUNT:
				self.outfile.write(str(polymer.monomer_count) + chr(9) \
				+ str(polymer.count) + chr(9) \
				+ str(log_freq_polymer_obs) + chr(9) \
				+ str(log_depart_from_pred) + '\n')

				if (log_depart_from_pred >= 0.3) & (self.zero_three_dep != TRUE):
					sys.stdout.write('0.3 departure for ' \
					+ str(polymer.monomer_name) + ': ' \
					+ str(polymer.monomer_count) + '\n')
					self.zero_three_dep = TRUE

				if (log_depart_from_pred >= 0.5) & (self.zero_five_dep != TRUE):
					sys.stdout.write('0.5 departure for ' \
					+ str(polymer.monomer_name) + ': ' \
					+ str(polymer.monomer_count) + '\n')
					self.zero_five_dep = TRUE

		else:
			sys.stdout.write('Div by 0 calculating departure for ' \
			+ str(polymer.name) + '\n')

	# End print_details


if __name__ == '__main__':
	Poly()

